#include <cuda_runtime.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


#define BLOCK_SIZE 1024

struct Body {
    float x, y, z, vx, vy, vz;
};

void min(double time, double& minT) {
    if (time < minT)
        minT = time;
}

void max(double time, double& maxT) {
    if (time > maxT)
        maxT = time;
}

void randomizeBodies(float* data, int n) {
    for (int i = 0; i < n; i++) {
        data[i] = 2.0f * (rand() / (float)RAND_MAX) - 1.0f;
    }
}

__global__
void bodyForce(Body* p, float dt, int n) { //Glavna funkcija koju provodi svaki treth
    int i = blockDim.x * blockIdx.x + threadIdx.x; // Index thread
    if (i < n) {
        float Fx = 0.0f;
        float Fy = 0.0f;
        float Fz = 0.0f;

        for (int j = 0; j < n; j++) {
            float dx = p[j].x - p[i].x;
            float dy = p[j].y - p[i].y;
            float dz = p[j].z - p[i].z;
            float distSqr = dx * dx + dy * dy + dz * dz;
            float invDist = 1.0f / sqrtf(distSqr);
            float invDist3 = invDist * invDist * invDist;

            Fx += dx * invDist3;
            Fy += dy * invDist3;
            Fz += dz * invDist3;

        }

        p[i].vx += dt * Fx; p[i].vy += dt * Fy; p[i].vz += dt * Fz;
    }
}

int main(const int argc, const char** argv) {
    int device_count;
    cudaGetDeviceCount(&device_count);
    printf("Broj gpu-a: %d\n", device_count);
    for(int t = 0; t<device_count;t++){
    cudaDeviceProp device_prop;
    cudaGetDeviceProperties(&device_prop, t);
    printf("GPU %d - %s \n", t, device_prop.name);
    //std::cout<<t<<" - "<<device_prop.name<<std::endl;
    //cudaSetDevice(1);
    
    }
    printf("\n\n");
    for(int setgpu = 0; setgpu<device_count;setgpu++){
        cudaSetDevice(setgpu);
        int nBodies = 0;
        for (int r = 0; r < 3; r++) { // Provodimo 3 simulacije sa razlocitim brojem tijela 
            if (r == 0) {
                nBodies = 50000; //broj tijela
            }
            else if (r == 1) {
                nBodies = 150000; //broj tijela
            }
            else {
                nBodies = 350000; //broj tijela
            }
            //double minT = 10000.0, maxT = 0.0;
            if (argc > 1) {
                nBodies = atoi(argv[1]); //ako postoji,argument kod pokretanja programa postaje broj tijela
                r = 3;

            }

            const float dt = 0.01f; // time step
            const int iterations = 11;  // simulation iterations

            int allocBytes = nBodies * sizeof(Body); // Velicina potrebne alokacije
            float* buf = (float*)malloc(allocBytes); // Alokacija memorije u RAM-u (host)
            Body* p = (Body*)buf;  //pokazivac p postaje buf ????

            randomizeBodies(buf, 6 * nBodies); // Init pos / vel data

            float* d_buf; //Definiranje pokazivaca
            cudaMalloc(&d_buf, allocBytes); // Alokacija memorije na GPU
            Body* d_p = (Body*)d_buf;

            int grid = (nBodies + BLOCK_SIZE - 1) / BLOCK_SIZE; //Definiranje velicine GRID-a
            int MaxN = BLOCK_SIZE * grid; //Najveci broj niti sa ovim postavkama
            double timesum = 0.0; //definiramo varijablu za izracunavalje sveukukpnog vremena izvodenja

            printf("Number of body's: %d\nBlock size: %d\nGrid size: %d\nMaximum n in with this setup (Block * Grid): %d\nThreds not used: %d\nUsed threds: %d %%\nNumber of iterations: %d\n\n", nBodies, BLOCK_SIZE, grid, MaxN, MaxN - nBodies, nBodies * 100 / MaxN, iterations);
            cudaDeviceProp device_prop;
            cudaGetDeviceProperties(&device_prop, setgpu);
            printf("GPU %d - %s \n", setgpu, device_prop.name);
            for (int i = 1; i <= iterations; i++) {
                int start = clock(); // pocetak stopanja vremena

                cudaMemcpy(d_buf, buf, allocBytes, cudaMemcpyHostToDevice); //Kopiramo podatke iz radne memorije na memoriju GPU-a
                //hipLaunchKernelGGL(bodyForce, dim3(grid), dim3(BLOCK_SIZE), 0, 0, d_p, dt, nBodies); // compute interbody forces
                bodyForce << <grid, BLOCK_SIZE >> > (d_p, dt, nBodies);
                cudaMemcpy(buf, d_buf, allocBytes, cudaMemcpyDeviceToHost);//Kopiramo podatke sa memorije GPU-a na memoriju hosta

                for (int j = 0; j < nBodies; j++) { // integrate position
                    p[j].x += p[j].vx * dt;
                    p[j].y += p[j].vy * dt;
                    p[j].z += p[j].vz * dt;
                }

                const double timediff = (clock() - start) / 1000000.0; // Vrijeme provodenja operacije
                //min(timediff, minT);//Pozivanje funkcije min()
                //max(timediff, maxT);//Pozivanje funkcije max()
                if (i > 1) { // First iter is warm up
                    timesum += timediff; //Sumiranje vremena svih operacija
                }

                //printf("Iteration %d: %.5f seconds\n", i, timediff);
                printf("%.5f\n", timediff);

            }
            double avgT = timesum / (double)(iterations - 1); //prosjecno vrijeme iteracije 


            //printf("\nMinimum time: %0.5f seconds, Maximum time: %0.5f seconds, Average time: %0.5f\n", minT, maxT, avgT);
            printf("%d Bodies: average %0.0f Operations / second\n\n\n", nBodies, round(nBodies / timesum));

            free(buf); // Release host memory
            cudaFree(d_buf); // Release device memory
        }
    }
}
