#include "hip/hip_runtime.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define BLOCK_SIZE 1024

struct Body{
    float x, y, z, vx, vy, vz;
};

void min(double time, double &minT){
    if(time<minT)
        minT = time;
}

void max(double time, double &maxT){
    if(time>maxT)
        maxT = time;
}

void randBodies(float *data, int n) {
  for (int i = 0; i < n; i++) {
    data[i] = 2.0f * (rand() / (float)RAND_MAX) - 1.0f;
  }
}

__global__
void bodyForce(Body *p, float dt, int n) { //Glavna funkcija koju provodi svaki treth
  int i = hipBlockDim_x * hipBlockIdx_x + hipThreadIdx_x; // Index niti
  if (i < n) {
    float rx = 0.0f;
    float ry = 0.0f;
    float rz = 0.0f;  

    for (int j = 0; j < n; j++) {
      float dx = p[j].x - p[i].x; // udaljenost x
      float dy = p[j].y - p[i].y; // udaljenost y
      float dz = p[j].z - p[i].z; // udaljenost z

      float distSqr = dx*dx + dy*dy + dz*dz; // x^2 + y^2 + z^2
      
	    float invDistCube = 1.0f / sqrtf(distSqr); // sila = 1/udaljenost
      float invDist3 = invDistCube * invDistCube * invDistCube; 

      rx += dx * invDist3;
      ry += dy * invDist3;
      rz += dz * invDist3;
      
    }

    p[i].vx += dt*rx; //
    p[i].vy += dt*ry; //pomak u smjeru po vremenskoj jedinici
    p[i].vz += dt*rz; //
  }
}


int main(const int argc, const char** argv) {
    int nBodies = 0;
    for(int r = 0; r<3;r++){ // Provodimo 3 simulacije sa razlocitim brojem tijela 
        if(r == 0){
            nBodies = 50000; //broj tijela
        }
        else if (r == 1){
            nBodies = 150000; //broj tijela
        }
        else{
            nBodies = 350000; //broj tijela
        }
        double minT = 10000.0, maxT = 0.0;
        if (argc > 1){
            nBodies = atoi(argv[1]); //ako postoji,argument kod pokretanja programa postaje broj tijela
            r = 3;

        } 

        const float dt = 0.01f; // time step
        const int iterations = 10;  // simulation iterations

        int allocBytes = nBodies*sizeof(Body); // Velicina potrebne alokacije
        float *buf = (float*)malloc(allocBytes); // Alokacija memorije u RAM-u (host)
        Body *p = (Body*)buf;  //pokazivac p postaje buf 


        float *d_buf; //Definiranje pokazivaca
        hipMalloc(&d_buf, allocBytes); // Alokacija memorije na GPU
        Body *d_p = (Body*)d_buf;

        randBodies(buf, 6*nBodies); 

        int grid = (nBodies + BLOCK_SIZE - 1) / BLOCK_SIZE; //Definiranje velicine GRID-a
        int maxN = BLOCK_SIZE*grid; //Najveci broj niti sa ovim postavkama
        double timesum = 0.0; //definiramo varijablu za izracunavalje sveukukpnog vremena izvodenja

        printf("Number of body's: %d\nBlock size: %d\nGrid size: %d\nMaximum n in with this setup (Block * Grid): %d\nThreds not used: %d\nUsed threds: %d %%\nNumber of iterations: %d\n\n",nBodies, BLOCK_SIZE, grid, maxN, maxN-nBodies, nBodies*100/maxN, iterations); 
        for (int i = 1; i <= iterations; i++) {
            int start = clock(); // pocetak stopanja vremena

            hipMemcpy(d_buf, buf, allocBytes, hipMemcpyHostToDevice); //Kopiramo podatke iz radne memorije na memoriju GPU-a
            hipLaunchKernelGGL(bodyForce, dim3(grid), dim3(BLOCK_SIZE), 0, 0, d_p, dt, nBodies); // compute interbody forces
            hipMemcpy(buf, d_buf, allocBytes, hipMemcpyDeviceToHost);//Kopiramo podatke sa memorije GPU-a na memoriju hosta

            
            for (int j = 0 ; j < nBodies; j++) { // Izračunavamo poziciju svakog tijela
            p[j].x += p[j].vx*dt;
            p[j].y += p[j].vy*dt;
            p[j].z += p[j].vz*dt;
            }

            const double timediff = (clock()-start) / 1000000.0; // Vrijeme provodenja operacije
            min(timediff, minT);//Pozivanje funkcije min()
            max(timediff, maxT);//Pozivanje funkcije max()
            if (i > 1) { 
            timesum += timediff; //Sumiranje vremena svih operacija
            }

            printf("Iteration %d: %.5f seconds\n", i, timediff);


        }
        double avgT = timesum / (double)(iterations-1); //prosjecno vrijeme iteracije 

        
        printf("\nMinimum time: %0.5f seconds, Maximum time: %0.5f seconds, Average time: %0.5f\n",minT, maxT, avgT);
        printf("%d Bodies: average %0.0f Operations / second\n\n\n", nBodies, round(nBodies / timesum));

        free(buf); // Oslobađanje memorije domaćina
        hipFree(d_buf); // Oslobađanje memorije grafičke kartice
    }
}
